#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.eventbus
      @file: event.py
   @created: Wed, 30 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""


class Event:
    """TODO"""

    def __init__(self, _name_: str, **kwargs):
        self.name = _name_
        self.kwargs = kwargs

    def __str__(self):
        return f"{self.name}: {str(self.kwargs)}"

    def __repr__(self):
        return str(self)

    def __eq__(self, other: 'Event'):
        return self.name == other.name and self.kwargs == other.kwargs

    def __getitem__(self, item: str):
        return self.kwargs[item]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
      @file: app_extension.py
   @created: Wed, 30 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from hspylib.core.enums.enumeration import Enumeration


class AppExtension(Enumeration):
    """Available HSPyLib application extensions"""
    GRADLE = 'gradle'
    GIT = 'git'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.application
      @file: application.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""
import sys


class ExitHooks(object):
    """TODO"""

    def __init__(self):
        self._orig_exit = None
        self.exit_code = None
        self.exception = None

    def hook(self):
        self._orig_exit = sys.exit
        sys.exit = self.exit
        sys.excepthook = self.exc_handler

    def exit(self, code=0):
        self.exit_code = code
        self._orig_exit(code)

    def exc_handler(self, exc_type, exc, *args):
        self.exception = exc



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.icons.font_awesome
      @file: dashboard_icons.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""
from hspylib.modules.cli.icons.font_awesome.awesome import Awesome


class DashboardIcons(Awesome):
    """
        Dashboard UI icons.
        Codes can be found here:
        - https://fontawesome.com/cheatsheet?from=io
    """

    # @formatter:off
    SEARCH = '\uF002'        # 
    EMAIL = '\uF003'         # 
    LIKE = '\uF004'          # 
    USER = '\uF007'          # 
    MOVIE = '\uF008'         # 
    DASHBOARD = '\uF00A'     # 
    POWER = '\uF011'         # 
    SETTINGS = '\uF013'      # 
    ADJUSTMENTS = '\uFB2D'   # שּׂ
    HOME = '\uF015'          # 
    DOCUMENT = '\uF016'      # 
    HISTORY = '\uF017'       # 
    DELIVERED = '\uFB3D'     # ﬽
    READ = '\uFB3C'          # לּ
    DOWNLOAD = '\uF019'      # 
    ARCHIVE = '\uF01C'       # 
    PLAY = '\uF01D'          # 
    FLAG = '\uF024'          # 
    FAVORITE = '\uF02E'      # 
    PRINTER = '\uF02F'       # 
    CAMERA = '\uF030'        # 
    LIST = '\uF03A'          # 
    CODE = '\uFABF'          # 謹
    FORM = '\uF767'          # 
    FORMAT = '\uF9C5'        # 暈
    PICTURE = '\uF03E'       # 
    ERROR = '\uF06A'         # 
    ALERT = '\uF071'         # 
    AGENDA = '\uF073'        # 
    CHAT = '\uF075'          # 
    SHOPPING = '\uF07A'      # 
    GRAPHS = '\uF080'        # 
    FOLDER = '\uF74A'        # 
    FOLDER_OPEN = '\uF07C'   # 
    EXIT = '\uF08B'          # 
    SEND = '\uF989'          # 黎
    SENT = '\uF1D8'          # 
    COMPASS = '\uF124'       # 
    GRAPH = '\uF08F'         # 
    SAVE = '\uF0C7'          # 
    CLOUD_DOWN = '\uF0ED'    # 
    CLOUD_UP = '\uF0EE'      # 
    NOTIFICATION = '\uF0F3'  # 
    DATABASE = '\uF1C0'      # 
    PLUGIN = '\uF930'        # 擄
    PLUG_IN = '\uF1E6'       # 
    PLUG_OUT = '\uF492'      # 
    UNPLUGGED = '\uFBA4'     # ﮤ
    PLUGGED = '\uFBA3'       # ﮣ
    KILL = '\uFB8A'          # ﮊ
    IDENTITY = '\uF2BB'      # 
    MUSIC = '\uF3B5'         # 
    MUTE = '\uF86B'          # 
    UNMUTE = '\uF86C'        # 
    SECRET = '\uF49C'        # 
    VERIFIED = '\uFC8F'      # ﲏ
    LIGHT_ON = '\uF834'      # 
    LIGHT_OFF = '\uF835'     # 
    DOCKER = '\uF308'        # 
    IE = '\uF7FF'            # 
    GIT = '\uF813'           # 
    PYTHON = '\uF81F'        # 
    CPLUSPLUS = '\uFB70'     # ﭱ
    JAVA = '\uF4E4'          # 
    CLOUD_PUT = '\uF40A'     # 
    CLOUD_GET = '\uF409'     # 
    EXPORT = '\uF5F6'        # 
    IMPORT = '\uF5F9'        # 
    # @formatter:on


if __name__ == '__main__':
    DashboardIcons.demo_icons()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.mock
      @file: mock_server.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from http.server import HTTPServer
from random import randint
from threading import Thread
from time import sleep
from typing import Optional, Tuple

from requests.structures import CaseInsensitiveDict

from hspylib.core.enums.charset import Charset
from hspylib.core.enums.content_type import ContentType
from hspylib.core.enums.http_code import HttpCode
from hspylib.core.enums.http_method import HttpMethod
from hspylib.modules.fetch.http_response import HttpResponse
from hspylib.modules.mock.mock_server_handler import MockServerHandler


class MockServer(HTTPServer):
    """TODO"""

    RANDOM_PORT = randint(49152, 65535)

    class ServerThread(Thread):
        def __init__(self, parent: 'MockServer'):
            super().__init__()
            self.parent = parent

        def run(self) -> None:
            self.parent.serve_forever()

    class MockResponse(HttpResponse):
        """TODO"""

        def __init__(
            self,
            parent: 'MockServer',
            method: HttpMethod,
            url: str,
            status_code: HttpCode = None,
            body: str = None,
            headers=None,
            encoding: Charset = Charset.UTF_8,
            content_type=ContentType.APPLICATION_JSON):

            super().__init__(method, url, status_code, body, headers, encoding, content_type)
            self.parent = parent
            self.received_body = False

        def then_wait(self, delay: int) -> 'MockServer.MockResponse':
            sleep(delay)
            return self

        def then_return(
            self,
            code: HttpCode,
            body: str = None,
            headers=None,
            encoding: Charset = Charset.UTF_8,
            content_type=ContentType.APPLICATION_JSON) -> 'MockServer':
            """TODO"""

            response = self.parent.mock(self.method, self.url)
            response.status_code = code
            response.body = body
            response.headers = headers if headers else []
            response.encoding = encoding
            response.content_type = content_type
            if response.content_type:
                response.content_type.charset = encoding

            return self.parent

        def then_return_with_received_body(
            self,
            code: HttpCode,
            headers: CaseInsensitiveDict = None,
            encoding: Charset = Charset.UTF_8,
            content_type=ContentType.APPLICATION_JSON) -> 'MockServer':
            """TODO"""

            response = self.parent.mock(self.method, self.url)
            response.received_body = True
            response.body = None
            response.status_code = code
            response.headers = headers if headers else []
            response.encoding = encoding
            response.content_type = content_type
            if response.content_type:
                response.content_type.charset = encoding

            return self.parent

    def __init__(self, hostname: str, port: int):
        self._mocks = {}
        self.hostname = hostname
        self.port = port
        self.version = '0.9.0'
        super().__init__(self.address(), MockServerHandler)

    def mock(self, method: HttpMethod, url: str) -> Optional[MockResponse]:
        """TODO"""
        try:
            return self._mocks[method][url]
        except KeyError:
            return None

    def address(self) -> Tuple[str, int]:
        """TODO"""
        return self.hostname, self.port

    def is_allowed(self, method: HttpMethod) -> bool:
        """TODO"""
        return method in self._mocks

    def start(self) -> None:
        """TODO"""
        runner = self.ServerThread(self)
        runner.start()

    def stop(self) -> None:
        """TODO"""
        self.shutdown()
        self.server_close()

    def when_request(self, method: HttpMethod, url: str) -> Optional[MockResponse]:
        """TODO"""
        request = self.mock(method=method, url=url) or self.MockResponse(self, method, url)
        self._mocks[method] = self._mocks[method] if method in self._mocks else {}
        self._mocks[method][url] = request
        return request



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file

   @project: HSPyLib
   @package: main.addons.widman.widgets
      @file: widget_time_calc.py
   @created: Thu, 20 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import math
import re
from typing import List, Optional

from hspylib.addons.widman.widget import Widget
from hspylib.core.enums.exit_code import ExitCode
from hspylib.core.exception.exceptions import WidgetExecutionError
from hspylib.core.tools.commons import sysout
from hspylib.modules.cli.icons.font_awesome.widget_icons import WidgetIcons
from hspylib.modules.cli.tui.extra.minput.minput import MenuInput, minput


class WidgetTimeCalc(Widget):
    """HSPyLib Widget to calculate time based operations"""
    WIDGET_ICON = WidgetIcons.CLOCK
    WIDGET_NAME = "TimeCalc"
    TOOLTIP = "Calculate time based operations."
    USAGE = "Usage: TimeCalc [+d|++decimal] <HH1:MM1[:SS1]> <+|-> <HH2:MM2[:SS2]>"
    VERSION = (0, 1, 0)

    def __init__(self):
        super().__init__(
            self.WIDGET_ICON,
            self.WIDGET_NAME,
            self.TOOLTIP,
            self.USAGE,
            self.VERSION)

        self.total_seconds = 0
        self.op = '+'
        self.decimal = False
        self.args = None

    def parse_args(self, args: List[str]) -> Optional[ExitCode]:
        """TODO"""

        if (not args or len(args) < 3) or not any(a in args for a in ['+h', '++help']):
            if not self._read_args():
                return ExitCode.ERROR
        elif args[0] in ['+h', '++help']:
            sysout(self.usage())
            return ExitCode.SUCCESS
        elif args[0] in ['+v', '++version']:
            sysout(self.version())
            return ExitCode.SUCCESS
        elif args[0] in ['+d', '++decimal']:
            self.decimal = True
            args = args[1:]

        if not self.args:
            self.args = args

        return None

    def execute(self, args: List[str] = None) -> ExitCode:
        """TODO"""

        ret_val = self.parse_args(args)

        if ret_val is not None:
            return ret_val

        for tm in self.args:
            if re.match(r"[+-]", tm):
                self.op = tm
            elif re.match(r"^([0-9]{1,2}:?)+", tm):
                try:
                    parts = [int(math.floor(float(s))) for s in tm.split(':')]
                except ValueError:
                    return ExitCode.ERROR
                f_hours = parts[0] if len(parts) > 0 else 0
                f_minutes = parts[1] if len(parts) > 1 else 0
                f_secs = parts[2] if len(parts) > 2 else 0
                tm_amount = ((f_hours * 60 + f_minutes) * 60 + f_secs)

                if self.op == '+':
                    self.total_seconds += tm_amount
                elif self.op == '-':
                    self.total_seconds -= tm_amount
            else:
                raise WidgetExecutionError(f"Invalid time input: '{tm}'")

        self.total_seconds, seconds = divmod(self.total_seconds, 60)
        hours, minutes = divmod(self.total_seconds, 60)

        sysout('%HOM%%ED2%%MOD(0)%', end='')
        if self.decimal:
            sysout(f"{hours:02d}.{self._decimal(minutes):02d}.{self._decimal(seconds):02d}")
        else:
            sysout(f"{hours:02d}:{self._decimal(minutes):02d}:{self._decimal(seconds):02d}")

        return ExitCode.SUCCESS

    def _decimal(self, time_raw: int = 0) -> int:
        """ Convert a raw time into decimal """
        return int(round(((time_raw / 60.00) * 100.00) if self.decimal else time_raw))

    def _read_args(self) -> bool:
        """ When no input is provided (e.g:. when executed from dashboard). Prompt the user for the info. """
        # @formatter:off
        form_fields = MenuInput.builder() \
            .field() \
                .label('Time 1') \
                .itype('masked') \
                .value('|##:##:##') \
                .build() \
            .field() \
                .label('Operation') \
                .itype('select') \
                .value('+|-') \
                .build() \
            .field() \
                .label('Time 2') \
                .itype('masked') \
                .value('|##:##:##') \
                .build() \
            .build()
        # @formatter:on

        result = minput(form_fields)
        self.args = [value for _, value in result.__dict__.items()] if result else []

        return bool(result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.config
      @file: properties.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import logging as log
import os
import re
from configparser import ConfigParser
from typing import List, Optional

import yaml

from hspylib.core.tools.commons import flatten_dict, run_dir, str_to_bool, touch_file


class Properties:
    """TODO"""

    _default_name = 'application'
    _default_extension = '.properties'
    _profiled_format = '{}-{}{}'
    _simple_format = '{}{}'

    @staticmethod
    def environ_name(property_name: str) -> str:
        """Retrieve the environment name of the specified property name
        :param property_name: the name of the property using space, dot or dash notations
        """
        return re.sub('[ -.]', '_', property_name).upper()

    @staticmethod
    def read_properties(all_lines: List[str]) -> dict:
        return {
            p[0].strip(): p[1].strip() for p in [
                p.split('=', 1) for p in list(
                    filter(lambda l: re.match('[a-zA-Z0-9][._\\-a-zA-Z0-9]* *=.*', l), all_lines)
                )
            ]
        }

    @staticmethod
    def read_cfg_or_ini(all_lines: List[str]):
        string = os.linesep.join(all_lines)
        all_cfgs = {}
        cfg = ConfigParser()
        cfg.read_string(string)
        for section in cfg.sections():
            all_cfgs.update(dict(cfg.items(section)))
        return all_cfgs

    def __init__(self, filename: str = None, profile: str = None, load_dir: str = None):

        filename, extension = os.path.splitext(
            filename if filename else f'{self._default_name}{self._default_extension}')

        self.filename = filename
        self.extension = extension
        self.profile = profile if profile else os.environ.get('ACTIVE_PROFILE')
        self.load_dir = load_dir if load_dir else f'{run_dir()}/resources'
        self.filepath = None
        self.properties = {}
        self._read()

    def __str__(self):
        str_val = ''
        for key, value in self.properties.items():
            str_val += '{}{}={}'.format('\n' if str_val else '', key, value)
        return str_val

    def __getitem__(self, item: str):
        return self.get(item)

    def __iter__(self):
        return self.properties.__iter__()

    def __len__(self) -> int:
        """Retrieve the amount of properties"""
        return len(self.properties) if self.properties else 0

    def get(self, key: str, default_value=None) -> Optional[str]:
        """Get a property value as string or default_value if the property was not found"""
        return self.properties[key.strip()] if key.strip() in self.properties else default_value

    def get_int(self, key: str, default_value=None) -> Optional[int]:
        """Get and convert a property value into int or return a default value if any error occurred"""
        try:
            return int(self.get(key))
        except TypeError:
            return default_value

    def get_float(self, key: str, default_value=None) -> Optional[float]:
        """Get and convert a property value into float or return a default value if any error occurred"""
        try:
            return float(self.get(key))
        except TypeError:
            return default_value

    def get_bool(self, key: str, default_value=None) -> Optional[bool]:
        """Get and convert a property value into bool or return a default value if any error occurred"""
        try:
            return str_to_bool(self.get(key).lower())
        except TypeError:
            return default_value

    def values(self) -> list:
        """Retrieve all values for all properties"""
        return list(self.properties.values())

    def _read(self) -> None:
        """Read all properties from the file"""
        self.filepath = self._find_path()
        if os.path.exists(self.filepath):
            self._parse()
        else:
            raise FileNotFoundError(
                f'File "{self.filepath}" does not exist')

    def _find_path(self) -> str:
        """Find the proper path for the properties file"""
        if self.profile:
            filepath = self._profiled_format \
                .format(self.filename, self.profile, self.extension)
        else:
            filepath = self._simple_format \
                .format(self.filename, self.extension)
        return f'{self.load_dir}/{filepath}'

    def _parse(self) -> None:
        """Parse the properties file according to it's extension"""
        if not os.path.isfile(self._find_path()):
            touch_file(self._find_path())
        with open(self.filepath, encoding='utf-8') as fh_props:
            if self.extension in ['.ini', '.cfg']:
                all_lines = list(map(str.strip, filter(None, fh_props.readlines())))
                self.properties.update(self.read_cfg_or_ini(all_lines))
            elif self.extension == '.properties':
                all_lines = list(map(str.strip, filter(None, fh_props.readlines())))
                self.properties.update(self.read_properties(all_lines))
            elif self.extension in ['.yml', '.yaml']:
                self.properties.update(flatten_dict(yaml.safe_load(fh_props)))
            else:
                raise NotImplementedError(f'Extension {self.extension} is not supported')
        log.info('Successfully loaded %d properties from:\n\t=>%s', len(self.properties), self.filepath)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.tui.extra
      @file: mdashboard.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from typing import List, Optional

from hspylib.core.tools.commons import sysout
from hspylib.core.tools.preconditions import check_state
from hspylib.modules.cli.keyboard import Keyboard
from hspylib.modules.cli.tui.extra.mdashboard.dashboard_builder import DashboardBuilder
from hspylib.modules.cli.tui.extra.mdashboard.dashboard_item import DashboardItem
from hspylib.modules.cli.vt100.vt_codes import vt_print
from hspylib.modules.cli.vt100.vt_colors import VtColors
from hspylib.modules.cli.vt100.vt_utils import prepare_render, restore_cursor, restore_terminal, set_enable_echo


def mdashboard(
    items: List[DashboardItem],
    items_per_line: int = 5,
    title: str = 'Please select one item',
    title_color: VtColors = VtColors.ORANGE,
    nav_color: VtColors = VtColors.YELLOW) -> Optional[DashboardItem]:
    """TODO"""

    return MenuDashBoard(items, items_per_line).dashboard(title, title_color, nav_color)


class MenuDashBoard:
    """TODO"""

    ICN = 'X'

    CELL_TPL = [
        [' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '],
        [' ', ' ', ' ', ICN, ' ', ' ', ' ', ' '],
        [' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '],
    ]

    SEL_CELL_TPL = [
        [' ', '┏', '━', ' ', ' ', '━', '┓', ' '],
        [' ', ' ', ' ', ICN, ' ', ' ', ' ', ' '],
        [' ', '┗', '━', ' ', ' ', '━', '┛', ' '],
    ]

    NAV_ICONS = '\u2190\u2191\u2192\u2193'
    NAV_FMT = "\n{}[Enter] Select  [{}] Navigate  [Tab] Next  [Esc] Quit %EL0%"

    @classmethod
    def builder(cls):
        """TODO"""

        return DashboardBuilder()

    def __init__(
        self,
        items: List[DashboardItem],
        items_per_line: int = 5):

        self.items = items
        self.done = None
        self.re_render = True
        self.tab_index = 0
        self.items_per_line = items_per_line
        check_state(
            len(self.CELL_TPL) == len(self.SEL_CELL_TPL) and len(self.CELL_TPL[0]) == len(self.SEL_CELL_TPL[0]),
            'Invalid CELL definitions. Selected and Unselected matrices should have the same lengths.')

    def dashboard(
        self,
        title: str,
        title_color: VtColors,
        nav_color: VtColors) -> Optional[DashboardItem]:
        """TODO"""

        ret_val = Keyboard.VK_NONE
        length = len(self.items)

        if length == 0:
            return None

        prepare_render(title, title_color)

        # Wait for user interaction
        while not self.done and ret_val not in [Keyboard.VK_ENTER, Keyboard.VK_ESC]:
            # Menu Renderization
            if self.re_render:
                self._render(nav_color)

            # Navigation input
            ret_val = self._nav_input()

        restore_terminal()
        selected = self.items[self.tab_index] if ret_val == Keyboard.VK_ENTER else None

        if selected and selected.on_trigger:
            selected.on_trigger()

        return selected

    def _render(self, nav_color: VtColors) -> None:
        """TODO"""

        restore_cursor()
        set_enable_echo()

        for idx, item in enumerate(self.items):
            self._print_cell(idx, item, MenuDashBoard.CELL_TPL if self.tab_index != idx else MenuDashBoard.SEL_CELL_TPL)

        sysout(f'%EL2%\r> %GREEN%{self.items[self.tab_index].tooltip}%NC%')
        sysout(self.NAV_FMT.format(nav_color.placeholder(), self.NAV_ICONS), end='')
        self.re_render = False

    def _print_cell(self, idx: int, item: DashboardItem, cell_template: List[List[str]]) -> None:
        """TODO"""

        num_cols, num_rows = len(cell_template[0]), len(cell_template)

        for row in range(0, num_rows):
            for col in range(0, num_cols):
                vt_print(f'{item.icon if cell_template[row][col] == self.ICN else cell_template[row][col]}')
            vt_print(f'%CUD(1)%%CUB({num_cols})%')
        if idx > 0 and (idx + 1) % self.items_per_line == 0:
            vt_print(f'%CUD(1)%%CUB({num_cols * self.items_per_line})%')  # Break the line
        elif idx + 1 < len(self.items):
            vt_print(f'%CUU({num_rows})%%CUF({num_cols})%')  # Continue with the same line

    def _nav_input(self) -> chr:
        """TODO"""

        length = len(self.items)
        keypress = Keyboard.read_keystroke()
        if keypress:
            if keypress == Keyboard.VK_ESC:
                self.done = True
            elif keypress == Keyboard.VK_UP:
                self.tab_index = max(0, self.tab_index - self.items_per_line)
            elif keypress == Keyboard.VK_DOWN:
                self.tab_index = min(length - 1, self.tab_index + self.items_per_line)
            elif keypress in [Keyboard.VK_LEFT, Keyboard.VK_SHIFT_TAB]:
                self.tab_index = max(0, self.tab_index - 1)
            elif keypress in [Keyboard.VK_RIGHT, Keyboard.VK_TAB]:
                self.tab_index = min(length - 1, self.tab_index + 1)
            elif keypress == Keyboard.VK_ENTER:
                pass  # Just exit

        self.re_render = True
        return keypress
